package linkedqueue;

public class LinkedQueue
{
    //protected final Object putLock = new Object();
    //protected LinkedNode head = new LinkedNode(null);
    protected LinkedNode last = head;
    protected int waitingForTake = 0;

    //@ instance invariant head != null;
    //@ instance invariant last != null;
    //@ instance invariant putLock != null;
    //@ instance invariant waitingForTake >= 0;
    //@ instance invariant \reach(head).has(last);
    /*@ behavior
      @     assignable head, last, putLock, waitingForTake;
      @     ensures \fresh(head, putLock) && head.next == null;
      @*/
    public LinkedQueue()
    {
    }

    /*@ behavior
      @    requires_redundantly x != null;
      @    ensures last.value == x && \fresh(last);
      @*/
    protected void insert(Object x)
    {
        synchronized (putLock)
        {
            //LinkedNode p = new LinkedNode(x);
            synchronized (last)
            {
            	atomicInsert(p);
            }
            if (waitingForTake > 0)
                putLock.notify();
            return;
        }
    }
    
    /*@ behavior
      @    requires n != null;
      @    assignable last, last.next;
      @*/
    protected void atomicInsert(LinkedNode n) {
		last.next = n;
		last = n;
    	
    }

    protected Object extract()
    {
        synchronized (this)
        {
            synchronized (head)
            {
            	return atomicExtract();
            }
        }
    }

    /*@ behavior
      @    assignable head, head.next.value;
      @    ensures \result == null || (\exists LinkedNode n; \old(\reach(head)).has(n); n.value == \result && !(\reach(head).has(n)));
      @*/
    protected Object atomicExtract() {
    	
		Object x = null;
		LinkedNode first = head.next;
		
		if (first != null)
		{
			x = first.value;
			first.value = null;
			head = first;
		}
		
		return x;
    	
    }

    /*@ behavior
       @    requires x != null;
       @    ensures true;
       @also
       @ behavior
       @    requires x == null;
       @    signals (Exception e) e instanceof IllegalArgumentException;
       @*/
    public void put(Object x)
    {
        if (x == null)
            //throw new IllegalArgumentException();
        insert(x);
    }

    /*@ behavior
       @    when !(head.next == null);
       @    ensures \result != null;
       @    signals (Exception e) e instanceof RuntimeException;
       @*/
    public Object take()
    {
        Object x = extract();
        if (x != null)
            return x;
        else
        {
            synchronized (putLock)
            {
                try
                {
                    ++waitingForTake;
                    for (;;)
                    {
                        x = extract();
                        if (x != null)
                        {
                            --waitingForTake;
                            return x;
                        }
                        else
                        {
                            putLock.wait();
                        }
                    }
                }
                catch (InterruptedException ex)
                {
                    --waitingForTake;
                    putLock.notify();
                    // throw ex;
                    //throw new RuntimeException();
                }
            }
        }
    }

    /*@ behavior
      @    ensures true;
      @*/
    public Object peek()
    {
        synchronized (head)
        {
            LinkedNode first = head.next;
            if (first != null)
                return first.value;
            else
                return null;
        }
    }

    /*@ behavior
      @    ensures \result <==> head.next == null;
      @*/
    public boolean isEmpty()
    {
        synchronized (head)
        {
            return head.next == null;
        }
    }
}